# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Any, Optional, Tuple, Union

import torch

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import coil_combination_method, complex_mul
from atommic.collections.reconstruction.nn.rim_base import conv_layers, rim_utils, rnn_cells


class RIMBlock(torch.nn.Module):
    """RIMBlock is a block of Recurrent Inference Machines (RIMs) as presented in [Lonning19]_.

    References
    ----------
    .. [Lonning19] Lonning19 K, Putzky P, Sonke JJ, Reneman L, Caan MW, Welling M. Recurrent inference machines for
        reconstructing heterogeneous MRI data. Medical image analysis. 2019 Apr 1;53:64-78.

    """

    def __init__(
        self,
        recurrent_layer=None,
        conv_filters=None,
        conv_kernels=None,
        conv_dilations=None,
        conv_bias=None,
        recurrent_filters=None,
        recurrent_kernels=None,
        recurrent_dilations=None,
        recurrent_bias=None,
        depth: int = 2,
        time_steps: int = 8,
        conv_dim: int = 2,
        no_dc: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        dimensionality: int = 2,
        consecutive_slices: int = 1,
        coil_combination_method: str = "SENSE",
    ):
        """Inits :class:`RIMBlock`.

        Parameters
        ----------
        recurrent_layer : torch.nn.Module
            Type of the recurrent layer. It can be ``GRU``, ``MGU``, ``IndRNN``. Check ``rnn_cells`` for more details.
        conv_filters : list of int
            Number of filters in the convolutional layers.
        conv_kernels : list of int
            Kernel size of the convolutional layers.
        conv_dilations : list of int
            Dilation of the convolutional layers.
        conv_bias : list of bool
            Bias of the convolutional layers.
        recurrent_filters : list of int
            Number of filters in the recurrent layers.
        recurrent_kernels : list of int
            Kernel size of the recurrent layers.
        recurrent_dilations : list of int
            Dilation of the recurrent layers.
        recurrent_bias : list of bool
            Bias of the recurrent layers.
        depth : int
            Number of sequence of convolutional and recurrent layers. Default is ``2``.
        time_steps : int
            Number of recurrent time steps. Default is ``8``.
        conv_dim : int
            Dimension of the convolutional layers. Default is ``2``.
        no_dc : bool
            If ``True`` the DC component is not used. Default is ``True``.
        fft_centered : bool
            If ``True`` the FFT is centered. Default is ``False``.
        fft_normalization : str
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : tuple of int
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int
            Coil dimension of the input. Default is ``1``.
        dimensionality : int
            Dimensionality of the input. Default is ``2``.
        consecutive_slices : int
            Number of consecutive slices. Default is ``1``.
        coil_combination_method : str
            Coil combination method. Default is ``"SENSE"``.
        """
        super().__init__()

        self.input_size = depth * 2
        self.time_steps = time_steps

        self.layers = torch.nn.ModuleList()
        for (
            (conv_features, conv_k_size, conv_dilation, l_conv_bias, nonlinear),
            (rnn_features, rnn_k_size, rnn_dilation, rnn_bias, rnn_type),
        ) in zip(
            zip(conv_filters, conv_kernels, conv_dilations, conv_bias, ["relu", "relu", None]),
            zip(
                recurrent_filters,
                recurrent_kernels,
                recurrent_dilations,
                recurrent_bias,
                [recurrent_layer, recurrent_layer, None],
            ),
        ):
            conv_layer = None

            if conv_features != 0:
                conv_layer = conv_layers.ConvNonlinear(
                    self.input_size,
                    conv_features,
                    conv_dim=conv_dim,
                    kernel_size=conv_k_size,
                    dilation=conv_dilation,
                    bias=l_conv_bias,
                    nonlinear=nonlinear,
                )
                self.input_size = conv_features

            if rnn_features != 0 and rnn_type is not None:
                if rnn_type.upper() == "GRU":
                    rnn_type = rnn_cells.ConvGRUCell
                elif rnn_type.upper() == "MGU":
                    rnn_type = rnn_cells.ConvMGUCell
                elif rnn_type.upper() == "INDRNN":
                    rnn_type = rnn_cells.IndRNNCell
                else:
                    raise ValueError("Please specify a proper recurrent layer type.")

                rnn_layer = rnn_type(
                    self.input_size,
                    rnn_features,
                    conv_dim=conv_dim,
                    kernel_size=rnn_k_size,
                    dilation=rnn_dilation,
                    bias=rnn_bias,
                )

                self.input_size = rnn_features

                self.layers.append(conv_layers.ConvRNNStack(conv_layer, rnn_layer))

        self.final_layer = torch.nn.Sequential(conv_layer)

        self.recurrent_filters = recurrent_filters

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        self.no_dc = no_dc

        if not self.no_dc:
            self.dc_weight = torch.nn.Parameter(torch.ones(1))
            self.zero = torch.zeros(1, 1, 1, 1, 1)

        self.dimensionality = dimensionality
        self.consecutive_slices = consecutive_slices
        self.coil_combination_method = coil_combination_method

    def forward(
        self,
        y: torch.Tensor,
        masked_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        prediction: torch.Tensor = None,
        hx: torch.Tensor = None,
        sigma: float = 1.0,
        keep_prediction: bool = False,
    ) -> Tuple[Any, Union[list, torch.Tensor, None]]:
        """Forward pass of :class:`RIMBlock`.

        Parameters
        ----------
        y : torch.Tensor
            Predicted k-space. Shape: ``[batch, coils, height, width, 2]``.
        masked_kspace : torch.Tensor
            Subsampled k-space. Shape: ``[batch, coils, height, width, 2]``.
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape: ``[batch, coils, height, width, 2]``.
        mask : torch.Tensor
            Subsampling mask. Shape: ``[batch, coils, height, width, 2]``.
        prediction : torch.Tensor, optional
            Initial (zero-filled) prediction. Shape: ``[batch, coils, height, width, 2]``.
        hx : torch.Tensor, optional
            Initial prediction for the hidden state. Shape: ``[batch, coils, height, width, 2]``.
        sigma : float, optional
            Noise level. Default is ``1.0``.
        keep_prediction : bool, optional
            Whether to keep the prediction. Default is ``False``.

        Returns
        -------
        Tuple[Any, Union[list, torch.Tensor, None]]
            Reconstructed image and hidden states.
        """
        batch = masked_kspace.shape[0]
        if self.dimensionality == 3 or self.consecutive_slices > 1:
            # 3D pred.shape = [batch, slices, coils, height, width, 2] -> [batch * slices, coils, height, width, 2]
            slices = masked_kspace.shape[1]
            y = y.reshape([batch * slices, *y.shape[2:]])
            masked_kspace = masked_kspace.reshape([batch * slices, *masked_kspace.shape[2:]])
            mask = mask.reshape([batch * slices, *mask.shape[2:]])
            sensitivity_maps = sensitivity_maps.reshape([batch * slices, *sensitivity_maps.shape[2:]])
        else:
            # 2D pred.shape = [batch, coils, height, width, 2]
            slices = 1

        if hx is None or (not isinstance(hx, list) and hx.dim() < 3):
            hx = [
                masked_kspace.new_zeros((masked_kspace.size(0), f, *masked_kspace.size()[2:-1]))
                for f in self.recurrent_filters
                if f != 0
            ]

        if prediction is None or prediction.ndim < 3:
            prediction = (
                y
                if keep_prediction
                else coil_combination_method(
                    ifft2(y, self.fft_centered, self.fft_normalization, self.spatial_dims),
                    sensitivity_maps,
                    method=self.coil_combination_method,
                    dim=self.coil_dim,
                )
            )

        if (self.consecutive_slices > 1 or self.dimensionality == 3) and prediction.dim() == 5:
            prediction = prediction.reshape([batch * slices, *prediction.shape[2:]])

        predictions = []
        for _ in range(self.time_steps):
            log_likelihood_gradient_prediction = rim_utils.log_likelihood_gradient(
                prediction,
                masked_kspace,
                sensitivity_maps,
                mask,
                sigma,
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
                self.coil_dim,
            ).contiguous()

            if self.consecutive_slices > 1 or self.dimensionality == 3:
                log_likelihood_gradient_prediction = log_likelihood_gradient_prediction.view(
                    [
                        batch * slices,
                        4,
                        log_likelihood_gradient_prediction.shape[2],
                        log_likelihood_gradient_prediction.shape[3],
                    ]
                ).permute(1, 0, 2, 3)

            for h, convrnn in enumerate(self.layers):
                hx[h] = convrnn(log_likelihood_gradient_prediction, hx[h])
                if self.consecutive_slices > 1 or self.dimensionality == 3:
                    hx[h] = hx[h].squeeze(0)
                log_likelihood_gradient_prediction = hx[h]

            log_likelihood_gradient_prediction = self.final_layer(log_likelihood_gradient_prediction)

            if self.dimensionality == 2:
                log_likelihood_gradient_prediction = log_likelihood_gradient_prediction.permute(0, 2, 3, 1)
            elif self.dimensionality == 3:
                log_likelihood_gradient_prediction = log_likelihood_gradient_prediction.permute(1, 2, 3, 0)
                for h in range(len(hx)):  # pylint: disable=consider-using-enumerate
                    hx[h] = hx[h].permute(1, 0, 2, 3)

            prediction = prediction + log_likelihood_gradient_prediction

            predictions.append(prediction)

        if self.consecutive_slices > 1 or self.dimensionality == 3:
            for i, pred in enumerate(predictions):
                predictions[i] = pred.reshape([batch, slices, *pred.shape[1:]])

        if self.no_dc:
            return predictions, hx

        soft_dc = torch.where(mask, y - masked_kspace, self.zero.to(masked_kspace)) * self.dc_weight
        current_kspace = [
            masked_kspace
            - soft_dc
            - fft2(
                complex_mul(e.unsqueeze(self.coil_dim), sensitivity_maps),
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            for e in predictions
        ]

        return current_kspace, hx



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.reconstruction.nn.unet_base.unet_block import Unet
from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet

__all__ = ["SegmentationUNet"]


class SegmentationUNet(BaseSegmentationNet):
    """Implementation of the (2D) UNet for MRI segmentation, as presented in [Ronneberger2015]_.

    References
    ----------
    .. [Ronneberger2015] O. Ronneberger, P. Fischer, and Thomas Brox. U-net: Convolutional networks for biomedical
        image segmentation. In International Conference on Medical image computing and computer-assisted intervention,
        pages 234–241. Springer, 2015.

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        return Unet(
            in_chans=self.input_channels,
            out_chans=cfg.get("segmentation_module_output_channels", 2),
            chans=cfg.get("segmentation_module_channels", 64),
            num_pool_layers=cfg.get("segmentation_module_pooling_layers", 2),
            drop_prob=cfg.get("segmentation_module_dropout", 0.0),
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, complex_conj, complex_mul
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import Conv2d
from atommic.collections.reconstruction.nn.crossdomain_base.crossdomain_block import MultiCoil
from atommic.collections.reconstruction.nn.didn_base.didn_block import DIDN
from atommic.collections.reconstruction.nn.mwcnn_base.mwcnn_block import MWCNN
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["KIKINet"]


class KIKINet(BaseMRIReconstructionModel):
    """Based on KIKINet implementation. Modified to work with multi-coil k-space data, as presented in [Taejoon2018]_.

    References
    ----------
    .. [Taejoon2018] Eo, Taejoon, et al. “KIKI-Net: Cross-Domain Convolutional Neural Networks for Reconstructing
        Undersampled Magnetic Resonance Images.” Magnetic Resonance in Medicine, vol. 80, no. 5, Nov. 2018, pp.
        2188–201. PubMed, https://doi.org/10.1002/mrm.27201.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`KIKINet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.num_iter = cfg_dict.get("num_iter")
        self.no_dc = cfg_dict.get("no_dc")

        kspace_model_architecture = cfg_dict.get("kspace_model_architecture")

        if kspace_model_architecture == "CONV":
            kspace_model = Conv2d(
                in_channels=cfg_dict.get("kspace_in_channels", 2),
                out_channels=cfg_dict.get("kspace_out_channels", 2),
                hidden_channels=cfg_dict.get("kspace_conv_hidden_channels"),
                n_convs=cfg_dict.get("kspace_conv_n_convs"),
                batchnorm=cfg_dict.get("kspace_conv_batchnorm"),
            )
        elif kspace_model_architecture == "DIDN":
            kspace_model = DIDN(
                in_channels=cfg_dict.get("kspace_in_channels", 2),
                out_channels=cfg_dict.get("kspace_out_channels", 2),
                hidden_channels=cfg_dict.get("kspace_didn_hidden_channels"),
                num_dubs=cfg_dict.get("kspace_didn_num_dubs"),
                num_convs_recon=cfg_dict.get("kspace_didn_num_convs_recon"),
            )
        elif kspace_model_architecture in ["UNET", "NORMUNET"]:
            kspace_model = NormUnet(
                cfg_dict.get("kspace_unet_num_filters"),
                cfg_dict.get("kspace_unet_num_pool_layers"),
                in_chans=cfg_dict.get("kspace_in_channels", 2),
                out_chans=cfg_dict.get("kspace_out_channels", 2),
                drop_prob=cfg_dict.get("kspace_unet_dropout_probability"),
                padding_size=cfg_dict.get("kspace_unet_padding_size"),
                normalize=cfg_dict.get("kspace_unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "KIKINet is currently implemented for kspace_model_architecture == 'CONV' or 'DIDN' or 'UNet'."
                f"Got kspace_model_architecture == {kspace_model_architecture}."
            )

        image_model_architecture = cfg_dict.get("imspace_model_architecture")

        if image_model_architecture == "MWCNN":
            image_model = MWCNN(
                input_channels=cfg_dict.get("imspace_in_channels", 2),
                first_conv_hidden_channels=cfg_dict.get("image_mwcnn_hidden_channels"),
                num_scales=cfg_dict.get("image_mwcnn_num_scales"),
                bias=cfg_dict.get("image_mwcnn_bias"),
                batchnorm=cfg_dict.get("image_mwcnn_batchnorm"),
            )
        elif image_model_architecture in ["UNET", "NORMUNET"]:
            image_model = NormUnet(
                cfg_dict.get("imspace_unet_num_filters"),
                cfg_dict.get("imspace_unet_num_pool_layers"),
                in_chans=cfg_dict.get("imspace_in_channels", 2),
                out_chans=cfg_dict.get("imspace_out_channels", 2),
                drop_prob=cfg_dict.get("imspace_unet_dropout_probability"),
                padding_size=cfg_dict.get("imspace_unet_padding_size"),
                normalize=cfg_dict.get("imspace_unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "KIKINet is currently implemented only with image_model_architecture == 'MWCNN' or 'UNet'."
                f"Got {image_model_architecture}."
            )

        self.image_model_list = torch.nn.ModuleList([image_model] * self.num_iter)
        self.kspace_model_list = torch.nn.ModuleList([MultiCoil(kspace_model, coil_dim=1)] * self.num_iter)

        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`KIKINet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        kspace = y.clone()
        zero = torch.zeros(1, 1, 1, 1, 1).to(kspace)

        for idx in range(self.num_iter):
            soft_dc = torch.where(mask.bool(), kspace - y, zero) * self.dc_weight

            kspace = self.kspace_model_list[idx](kspace)
            if kspace.shape[-1] != 2:
                kspace = kspace.permute(0, 1, 3, 4, 2).to(y)
                # this is necessary, but why?
                kspace = torch.view_as_real(kspace[..., 0] + 1j * kspace[..., 1])

            image = complex_mul(
                ifft2(
                    kspace,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                complex_conj(sensitivity_maps),
            ).sum(self.coil_dim)
            image = self.image_model_list[idx](image.unsqueeze(self.coil_dim)).squeeze(self.coil_dim)

            if not self.no_dc:
                image = fft2(
                    complex_mul(image.unsqueeze(self.coil_dim), sensitivity_maps),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ).type(image.type())
                image = kspace - soft_dc - image
                image = complex_mul(
                    ifft2(
                        image,
                        centered=self.fft_centered,
                        normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                    ),
                    complex_conj(sensitivity_maps),
                ).sum(self.coil_dim)

            if idx < self.num_iter - 1:
                kspace = fft2(
                    complex_mul(image.unsqueeze(self.coil_dim), sensitivity_maps),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ).type(image.type())

        return check_stacked_complex(image)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/nn/lpd/lpd.py

import torch
from torch import nn


class DualNet(nn.Module):
    """Dual Network for Learned Primal Dual Network."""

    def __init__(self, num_dual, **kwargs):
        """Inits :class:`DualNet`.

        Parameters
        ----------
        num_dual : int
            Number of dual for LPD algorithm.
        """
        super().__init__()
        if kwargs.get("dual_architecture") is None:
            n_hidden = kwargs.get("n_hidden")
            if n_hidden is None:
                raise ValueError("n_hidden is required for DualNet")

            self.dual_block = nn.Sequential(
                *[
                    nn.Conv2d(2 * (num_dual + 2), n_hidden, kernel_size=3, padding=1),
                    nn.PReLU(),
                    nn.Conv2d(n_hidden, n_hidden, kernel_size=3, padding=1),
                    nn.PReLU(),
                    nn.Conv2d(n_hidden, 2 * num_dual, kernel_size=3, padding=1),
                ]
            )
        else:
            self.dual_block = kwargs.get("dual_architecture")

    @staticmethod
    def compute_model_per_coil(model: nn.Module, data: torch.Tensor) -> torch.Tensor:
        """Computes model per coil.

        Parameters
        ----------
        model : torch.nn.Module
            Model to be computed.
        data : torch.Tensor
            Input data.

        Returns
        -------
        torch.Tensor
            Multicoil output.
        """
        output = []
        for idx in range(data.size(1)):
            subselected_data = data.select(1, idx)
            output.append(model(subselected_data))
        output = torch.stack(output, dim=1)
        return output

    def forward(self, h: torch.Tensor, forward_f: torch.Tensor, g: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DualNet`."""
        inp = torch.cat([h, forward_f, g], dim=-1).permute(0, 1, 4, 2, 3)
        return self.compute_model_per_coil(self.dual_block, inp).permute(0, 1, 3, 4, 2)


class PrimalNet(nn.Module):
    """Primal Network for Learned Primal Dual Network."""

    def __init__(self, num_primal, **kwargs):
        """Inits :class:`PrimalNet`.

        Parameters
        ----------
        num_primal : int
            Number of primal for LPD algorithm.
        """
        super().__init__()

        if kwargs.get("primal_architecture") is None:
            n_hidden = kwargs.get("n_hidden")
            if n_hidden is None:
                raise ValueError("Missing argument n_hidden.")
            self.primal_block = nn.Sequential(
                *[
                    nn.Conv2d(2 * (num_primal + 1), n_hidden, kernel_size=3, padding=1),
                    nn.PReLU(),
                    nn.Conv2d(n_hidden, n_hidden, kernel_size=3, padding=1),
                    nn.PReLU(),
                    nn.Conv2d(n_hidden, 2 * num_primal, kernel_size=3, padding=1),
                ]
            )
        else:
            self.primal_block = kwargs.get("primal_architecture")

    def forward(self, f: torch.Tensor, backward_h: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`PrimalNet`.

        Parameters
        ----------
        f : torch.Tensor
            Forward function.
        backward_h : torch.Tensor
            Backward function.

        Returns
        -------
        torch.Tensor
            Primal function.
        """
        inp = torch.cat([f, backward_h], dim=-1).permute(0, 3, 1, 2)
        return self.primal_block(inp).permute(0, 2, 3, 1)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/vit.py

from typing import Sequence, Union

import torch
from torch import nn

from atommic.collections.segmentation.nn.vit_base.patchembedding import PatchEmbeddingBlock
from atommic.collections.segmentation.nn.vit_base.transformer_block import TransformerBlock

__all__ = ["ViT"]


class ViT(nn.Module):
    """Implementation of a Vision Transformer (ViT), as presented in [Dosovitskiy2020]_.

    ViT supports Torchscript but only works for Pytorch after 1.8.

    References
    ----------
    .. [Dosovitskiy2020] Dosovitskiy A, Beyer L, Kolesnikov A, Weissenborn D, Zhai X, Unterthiner T, Dehghani M,
        Minderer M, Heigold G, Gelly S, Uszkoreit J. An image is worth 16x16 words: Transformers for image recognition
        at scale. arXiv preprint arXiv:2010.11929. 2020 Oct 22.

    .. note::
        This is a wrapper for monai implementation of PatchEmbeddingBlock.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/vit.py
    """

    def __init__(
        self,
        in_channels: int,
        img_size: Union[Sequence[int], int],
        patch_size: Union[Sequence[int], int],
        hidden_size: int = 768,
        mlp_dim: int = 3072,
        num_layers: int = 12,
        num_heads: int = 12,
        pos_embed: str = "conv",
        classification: bool = False,
        num_classes: int = 2,
        dropout_rate: float = 0.0,
        spatial_dims: int = 3,
        post_activation="Tanh",
        qkv_bias: bool = False,
    ):
        """Inits :class:`ViT`.

        Parameters
        ----------
        in_channels : int
            Dimension of input channels.
        img_size : Union[Sequence[int], int]
            Dimension of input image.
        patch_size : Union[Sequence[int], int]
            Dimension of patch size.
        hidden_size : int
            Dimension of hidden layer. Default is ``768``.
        mlp_dim : int
            Dimension of MLP layer. Default is ``3072``.
        num_layers : int
            Number of transformer layers. Default is ``12``.
        num_heads : int
            Number of attention heads. Default is ``12``.
        pos_embed : str
            Position embedding layer type. Default is ``"conv"``.
        classification : bool
            Whether to add a classification head. Default is ``False``.
        dropout_rate : float, optional
            Faction of the input units to drop. Default is ``0.0``.
        spatial_dims : int, optional
            Number of spatial dimensions. Default is ``3``.
        post_activation : str, optional
            Post activation layer type. Default is ``"Tanh"``.
        qkv_bias : bool, optional
            Whether to add a bias to query, key, value. Default is ``False``.
        """
        super().__init__()

        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")

        if hidden_size % num_heads != 0:
            raise ValueError("hidden_size should be divisible by num_heads.")

        self.classification = classification
        self.patch_embedding = PatchEmbeddingBlock(
            in_channels=in_channels,
            img_size=img_size,
            patch_size=patch_size,
            hidden_size=hidden_size,
            num_heads=num_heads,
            pos_embed=pos_embed,
            dropout_rate=dropout_rate,
            spatial_dims=spatial_dims,
        )
        self.blocks = nn.ModuleList(
            [TransformerBlock(hidden_size, mlp_dim, num_heads, dropout_rate, qkv_bias) for i in range(num_layers)]
        )
        self.norm = nn.LayerNorm(hidden_size)
        if self.classification:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, hidden_size))
            if post_activation == "Tanh":
                self.classification_head = nn.Sequential(nn.Linear(hidden_size, num_classes), nn.Tanh())
            else:
                self.classification_head = nn.Linear(hidden_size, num_classes)

    def forward(self, x: torch.Tensor) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
        """Forward pass of :class:`ViT`."""
        x = self.patch_embedding(x)
        if hasattr(self, "cls_token"):
            cls_token = self.cls_token.expand(x.shape[0], -1, -1)
            x = torch.cat((cls_token, x), dim=1)
        hidden_states_out = []
        for blk in self.blocks:
            x = blk(x)
            hidden_states_out.append(x)
        x = self.norm(x)
        if hasattr(self, "classification_head"):
            x = self.classification_head(x[:, 0])
        return x, hidden_states_out



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import argparse

import pytorch_lightning as pl
import torch
from omegaconf import DictConfig, OmegaConf

from atommic.collections.multitask.rs.nn.idslr import IDSLR
from atommic.collections.multitask.rs.nn.idslr_unet import IDSLRUNet
from atommic.collections.multitask.rs.nn.mtlrs import MTLRS
from atommic.collections.multitask.rs.nn.recseg_unet import RecSegUNet
from atommic.collections.multitask.rs.nn.segnet import SegNet
from atommic.collections.multitask.rs.nn.seranet import SERANet
from atommic.collections.quantitative.nn.qcirim import qCIRIM
from atommic.collections.quantitative.nn.qvarnet import qVarNet
from atommic.collections.quantitative.nn.qzf import qZF
from atommic.collections.reconstruction.nn.ccnn import CascadeNet
from atommic.collections.reconstruction.nn.cirim import CIRIM
from atommic.collections.reconstruction.nn.crnn import CRNNet
from atommic.collections.reconstruction.nn.dunet import DUNet
from atommic.collections.reconstruction.nn.jointicnet import JointICNet
from atommic.collections.reconstruction.nn.kikinet import KIKINet
from atommic.collections.reconstruction.nn.lpd import LPDNet
from atommic.collections.reconstruction.nn.modl import MoDL
from atommic.collections.reconstruction.nn.multidomainnet import MultiDomainNet
from atommic.collections.reconstruction.nn.proximal_gradient import ProximalGradient
from atommic.collections.reconstruction.nn.recurrentvarnet import RecurrentVarNet
from atommic.collections.reconstruction.nn.unet import UNet
from atommic.collections.reconstruction.nn.varnet import VarNet
from atommic.collections.reconstruction.nn.vsnet import VSNet
from atommic.collections.reconstruction.nn.xpdnet import XPDNet
from atommic.collections.reconstruction.nn.zf import ZF
from atommic.collections.segmentation.nn.attentionunet import SegmentationAttentionUNet
from atommic.collections.segmentation.nn.dynunet import SegmentationDYNUNet
from atommic.collections.segmentation.nn.lambdaunet import SegmentationLambdaUNet
from atommic.collections.segmentation.nn.unet import SegmentationUNet
from atommic.collections.segmentation.nn.unet3d import Segmentation3DUNet
from atommic.collections.segmentation.nn.unetr import SegmentationUNetR
from atommic.collections.segmentation.nn.vnet import SegmentationVNet
from atommic.core.conf.hydra_runner import hydra_runner
from atommic.utils import logging
from atommic.utils.exp_manager import exp_manager


def register_cli_subcommand(parser: argparse._SubParsersAction):
    """Register parser for the launch command."""
    parser_launch = parser.add_parser(
        "run",
        help="Launch atommic through cli given a configuration (yaml) file, e.g. atommic run -c /path/to/config.yaml",
    )
    parser_launch.add_argument(
        "-c",
        "--config-path",
        required=True,
        type=str,
        help="Path to the configuration file.",
    )
    parser_launch.add_argument(
        "-m",
        "--multi-run",
        action="store_true",
        help="Hydra Multi-Run for hyperparameter optimization.",
    )
    parser_launch.set_defaults(func=main)


@hydra_runner(config_path="../src", config_name="config")
def main(cfg: DictConfig):  # noqa: MC0001
    """
    Main function for training and running a model

    Parameters
    ----------
    cfg : Configuration (yaml) file.
        DictConfig
    """
    cfg = OmegaConf.load(f"{cfg.config_path}")

    logging.info(f"Config: {OmegaConf.to_yaml(cfg)}")

    trainer = pl.Trainer(**cfg.trainer)
    exp_manager(trainer, cfg.get("exp_manager", None))

    model_name = (cfg.model["model_name"]).upper()

    if model_name == "CASCADENET":
        model = CascadeNet(cfg.model, trainer=trainer)
    elif model_name == "CIRIM":
        model = CIRIM(cfg.model, trainer=trainer)
    elif model_name == "CRNNET":
        model = CRNNet(cfg.model, trainer=trainer)
    elif model_name == "DUNET":
        model = DUNet(cfg.model, trainer=trainer)
    elif model_name in ("E2EVN", "VN"):
        model = VarNet(cfg.model, trainer=trainer)
    elif model_name == "IDSLR":
        model = IDSLR(cfg.model, trainer=trainer)
    elif model_name == "IDSLRUNET":
        model = IDSLRUNet(cfg.model, trainer=trainer)
    elif model_name == "JOINTICNET":
        model = JointICNet(cfg.model, trainer=trainer)
    elif model_name == "KIKINET":
        model = KIKINet(cfg.model, trainer=trainer)
    elif model_name == "LPDNET":
        model = LPDNet(cfg.model, trainer=trainer)
    elif model_name == "MODL":
        model = MoDL(cfg.model, trainer=trainer)
    elif model_name == "MTLRS":
        model = MTLRS(cfg.model, trainer=trainer)
    elif model_name == "MULTIDOMAINNET":
        model = MultiDomainNet(cfg.model, trainer=trainer)
    elif model_name == "PROXIMALGRADIENT":
        model = ProximalGradient(cfg.model, trainer=trainer)
    elif model_name == "QCIRIM":
        model = qCIRIM(cfg.model, trainer=trainer)
    elif model_name == "QVN":
        model = qVarNet(cfg.model, trainer=trainer)
    elif model_name == "QZF":
        model = qZF(cfg.model, trainer=trainer)
    elif model_name == "RECSEGNET":
        model = RecSegUNet(cfg.model, trainer=trainer)
    elif model_name == "RVN":
        model = RecurrentVarNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONATTENTIONUNET":
        model = SegmentationAttentionUNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONDYNUNET":
        model = SegmentationDYNUNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONLAMBDAUNET":
        model = SegmentationLambdaUNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONUNET":
        model = SegmentationUNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONUNETR":
        model = SegmentationUNetR(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATION3DUNET":
        model = Segmentation3DUNet(cfg.model, trainer=trainer)
    elif model_name == "SEGMENTATIONVNET":
        model = SegmentationVNet(cfg.model, trainer=trainer)
    elif model_name == "SEGNET":
        model = SegNet(cfg.model, trainer=trainer)
    elif model_name == "SERANET":
        model = SERANet(cfg.model, trainer=trainer)
    elif model_name == "UNET":
        model = UNet(cfg.model, trainer=trainer)
    elif model_name == "VSNET":
        model = VSNet(cfg.model, trainer=trainer)
    elif model_name == "XPDNET":
        model = XPDNet(cfg.model, trainer=trainer)
    elif model_name == "ZF":
        model = ZF(cfg.model, trainer=trainer)
    else:
        raise NotImplementedError(f"{model_name} is not implemented in atommic.")

    if cfg.get("pretrained", None):
        checkpoint = cfg.get("checkpoint", None)
        logging.info(f"Loading pretrained model from {checkpoint}")

        # instantiate model
        if checkpoint.endswith(".atommic"):
            if "huggingface" in checkpoint:
                _, state_dict = model.from_pretrained(checkpoint)
            else:
                _, state_dict = model.restore_from(checkpoint)
        else:
            state_dict = torch.load(checkpoint, map_location="cpu")["state_dict"]

        model.load_state_dict(state_dict)

    if cfg.get("mode", None) == "train":
        logging.info("Validating")
        trainer.validate(model)
        logging.info("Training")
        trainer.fit(model)
    else:
        logging.info("Testing")
        trainer.test(model)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py

import torch
from torch import nn


class LUConv(nn.Module):
    """LU Convolutional Block.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(self, channels: int, act: nn.Module = nn.ELU, bias: bool = False):
        """Inits :class:`LUConv`.

        Parameters
        ----------
        channels : int
            Number of channels.
        act : nn.Module
            Activation function.
        bias : bool
            Whether to use bias.
        """
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(channels, channels, kernel_size=5, padding=2, bias=bias),
            nn.BatchNorm2d(channels),
            act(inplace=True),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`LUConv`."""
        return self.layers(x)


def _make_nconv(channels: int, depth: int, act: nn.Module = nn.ELU, bias: bool = False):
    """Make a stack of LUConv layers.

    Parameters
    ----------
    channels : int
        Number of channels.
    depth : int
        Number of LUConv layers.
    act : nn.Module
        Activation function.
    bias : bool
        Whether to use bias.

    Returns
    -------
    layers : nn.Sequential
        Stack of LUConv layers.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """
    return nn.Sequential(*[LUConv(channels=channels, act=act, bias=bias) for _ in range(depth)])


class InputTransition(nn.Module):
    """Input Transition Block.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int = 16,
        act: nn.Module = nn.ELU,
        bias: bool = False,
    ):
        """Inits :class:`InputTransition`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        act : nn.Module
            Activation function.
        bias : bool
            Whether to use bias.
        """
        super().__init__()

        if out_channels % in_channels != 0:
            raise ValueError(f"16 should be divisible by in_channels, got in_channels={in_channels}.")

        self.in_channels = in_channels
        self.act_function = act(inplace=True)
        self.conv_block = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=5, padding=2, bias=bias),
            nn.BatchNorm2d(out_channels),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`InputTransition`."""
        return self.act_function(self.conv_block(x) + x.repeat(1, 16 // self.in_channels, 1, 1))


class DownTransition(nn.Module):
    """Down Transition Block.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(
        self,
        in_channels: int,
        convs: int,
        act: nn.Module = nn.ELU,
        dropout_prob: float = 0.0,
        bias: bool = False,
    ):
        """Inits :class:`DownTransition`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        convs : int
            Number of LUConv layers.
        act : nn.Module
            Activation function. Default is ``ELU``.
        dropout_prob : float
            Dropout probability. Default is ``0.0``.
        bias : bool
            Whether to use bias. Default is ``False``.
        """
        super().__init__()

        out_channels = 2 * in_channels
        self.down_conv = nn.Conv2d(in_channels, out_channels, kernel_size=2, stride=2, bias=bias)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.act_function1 = act(inplace=True)
        self.act_function2 = act(inplace=True)
        self.ops = _make_nconv(out_channels, convs, act, bias)
        self.dropout = nn.Dropout2d(dropout_prob) if dropout_prob > 0.0 else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DownTransition`."""
        down = self.act_function1(self.bn1(self.down_conv(x)))
        return self.act_function2(self.ops(self.dropout(down) if self.dropout is not None else down) + down)


class UpTransition(nn.Module):
    """Up Transition Block.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        convs: int,
        act: nn.Module = nn.ELU,
        dropout_prob: float = 0.0,
    ):
        """Inits :class:`UpTransition`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        convs : int
            Number of LUConv layers.
        act : nn.Module
            Activation function. Default is ``ELU``.
        dropout_prob : float
            Dropout probability. Default is ``0.0``.
        """
        super().__init__()
        self.up_conv = nn.ConvTranspose2d(in_channels, out_channels // 2, kernel_size=2, stride=2)
        self.bn1 = nn.BatchNorm2d(out_channels // 2)
        self.dropout = nn.Dropout2d(dropout_prob) if dropout_prob > 0.0 else None
        self.dropout2 = nn.Dropout2d(0.5)
        self.act_function1 = act(inplace=True)
        self.act_function2 = act(inplace=True)
        self.ops = _make_nconv(out_channels, convs, act)

    def forward(self, x: torch.Tensor, skipx: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UpTransition`."""
        out = self.dropout(x) if self.dropout is not None else x
        xcat = torch.cat((self.act_function1(self.bn1(self.up_conv(out))), self.dropout2(skipx)), 1)
        return self.act_function2(self.ops(xcat) + xcat)


class OutputTransition(nn.Module):
    """Output Transition Block.

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act: nn.Module = nn.ELU,
        bias: bool = False,
    ):
        """Inits :class:`OutputTransition`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        act : nn.Module
            Activation function. Default is ``ELU``.
        bias : bool
            Whether to use bias. Default is ``False``.
        """
        super().__init__()

        self.conv_block = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=5, padding=2, bias=bias),
            nn.BatchNorm2d(out_channels),
            act(inplace=True),
        )
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`OutputTransition`."""
        # convolve 32 down to 2 channels
        return self.conv2(self.conv_block(x))


class VNet(nn.Module):
    """Implementation of the V-Net for MRI segmentation, as presented in [Milletari2016]_.

    References
    ----------
    .. [Milletari2016] Fausto Milletari, Nassir Navab, Seyed-Ahmad Ahmadi. V-Net: Fully Convolutional Neural Networks
        for Volumetric Medical Image Segmentation, 2016. https://arxiv.org/abs/1606.04797

    .. note::
        This is a wrapper for Vnet implementation.
        See: https://github.com/black0017/MedicalZooPytorch/blob/master/lib/medzoo/Vnet.py
    """

    def __init__(
        self,
        in_chans: int = 1,
        out_chans: int = 1,
        act: str = "elu",
        drop_prob: float = 0.5,
        bias: bool = False,
    ):
        """Inits :class:`VNet`.

        Parameters
        ----------
        in_chans : int
            Number of input channels. Default is ``1``.
        out_chans : int
            Number of output channels. Default is ``1``.
        act : nn.Module
            Activation function. Default is ``ELU``.
        drop_prob : float
            Dropout probability. Default is ``0.5``.
        bias : bool
            Whether to use bias. Default is ``False``.
        """
        super().__init__()

        if act == "elu":
            act = nn.ELU
        elif act == "relu":
            act = nn.ReLU
        elif act == "prelu":
            act = nn.PReLU
        elif act == "leakyrelu":
            act = nn.LeakyReLU
        else:
            raise ValueError(
                f"Activation function {act} not supported. Please choose between ReLU, PReLU, LeakyReLU, ELU."
            )

        self.in_tr = InputTransition(in_chans, 16, act, bias=bias)
        self.down_tr32 = DownTransition(16, 1, act, bias=bias)
        self.down_tr64 = DownTransition(32, 2, act, bias=bias)
        self.down_tr128 = DownTransition(64, 3, act, dropout_prob=drop_prob, bias=bias)
        self.down_tr256 = DownTransition(128, 2, act, dropout_prob=drop_prob, bias=bias)
        self.up_tr256 = UpTransition(256, 256, 2, act, dropout_prob=drop_prob)
        self.up_tr128 = UpTransition(256, 128, 2, act, dropout_prob=drop_prob)
        self.up_tr64 = UpTransition(128, 64, 1, act)
        self.up_tr32 = UpTransition(64, 32, 1, act)
        self.out_tr = OutputTransition(32, out_chans, act, bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`VNet`."""
        out16 = self.in_tr(x)
        out32 = self.down_tr32(out16)
        out64 = self.down_tr64(out32)
        out128 = self.down_tr128(out64)
        out256 = self.down_tr256(out128)
        x = self.up_tr256(out256, out128)
        x = self.up_tr128(x, out64)
        x = self.up_tr64(x, out32)
        x = self.up_tr32(x, out16)
        x = self.out_tr(x)
        return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import nn

from atommic.collections.common.parts.utils import check_stacked_complex
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.modl_base.modl_block import ConjugateGradient
from atommic.core.classes.common import typecheck

__all__ = ["ProximalGradient"]


class ProximalGradient(BaseMRIReconstructionModel):
    """Implementation of the Proximal/Conjugate Gradient, according to [Aggarwal2018]_, [Yaman2020]_.

    References
    ----------
    .. [Aggarwal2018] MoDL: Model Based Deep Learning Architecture for Inverse Problems by H.K. Aggarwal, M.P Mani, and
        Mathews Jacob in IEEE Transactions on Medical Imaging, 2018

    .. [Yaman2020] Yaman, B, Hosseini, SAH, Moeller, S, Ellermann, J, Uğurbil, K, Akçakaya, M. Self-supervised
        learning of physics-guided reconstruction neural networks without fully sampled reference data. Magn Reson
        Med. 2020; 84: 3172– 3191. https://doi.org/10.1002/mrm.28378

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`ProximalGradient`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)
        cfg_dict = OmegaConf.to_container(cfg, resolve=True)
        self.mu = nn.Parameter(torch.Tensor([cfg_dict.get("penalization_weight")]), requires_grad=True)
        self.dc_block = ConjugateGradient(
            cfg_dict.get("conjugate_gradient_iterations", 10),
            self.mu,
            self.fft_centered,
            self.fft_normalization,
            self.spatial_dims,
            self.coil_dim,
            self.coil_combination_method,
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,  # pylint: disable=unused-argument
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`ProximalGradient`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        return check_stacked_complex(self.dc_block(initial_prediction, sensitivity_maps, mask))



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.nn.attentionunet import SegmentationAttentionUNet  # noqa: F401
from atommic.collections.segmentation.nn.dynunet import SegmentationDYNUNet  # noqa: F401
from atommic.collections.segmentation.nn.lambdaunet import SegmentationLambdaUNet  # noqa: F401
from atommic.collections.segmentation.nn.unet import SegmentationUNet  # noqa: F401
from atommic.collections.segmentation.nn.unet3d import Segmentation3DUNet  # noqa: F401
from atommic.collections.segmentation.nn.unetr import SegmentationUNetR  # noqa: F401
from atommic.collections.segmentation.nn.vnet import SegmentationVNet  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import CascadeNetBlock, Conv2d
from atommic.core.classes.common import typecheck

__all__ = ["CascadeNet"]


class CascadeNet(BaseMRIReconstructionModel):
    """Implementation of the Deep Cascade of Convolutional Neural Networks, as presented in [Schlemper2017]_.

    References
    ----------
    .. [Schlemper2017] Schlemper, J., Caballero, J., Hajnal, J. V., Price, A., & Rueckert, D., A Deep Cascade of
        Convolutional Neural Networks for MR Image Reconstruction. Information Processing in Medical Imaging (IPMI),
        2017.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`CascadeNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        # Cascades of CascadeCNN blocks
        self.reconstruction_module = torch.nn.ModuleList(
            [
                CascadeNetBlock(
                    Conv2d(
                        in_channels=2,
                        out_channels=2,
                        hidden_channels=cfg_dict.get("hidden_channels"),
                        n_convs=cfg_dict.get("n_convs"),
                        batchnorm=cfg_dict.get("batchnorm"),
                    ),
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                    no_dc=cfg_dict.get("no_dc"),
                )
                for _ in range(cfg_dict.get("num_cascades"))
            ]
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`CascadeNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        prediction = y.clone()
        for cascade in self.reconstruction_module:
            prediction = cascade(prediction, y, sensitivity_maps, mask)
        return check_stacked_complex(
            coil_combination_method(
                ifft2(prediction, self.fft_centered, self.fft_normalization, self.spatial_dims),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:https://github.com/NKI-AI/direct/blob/main/direct/nn/multidomainnet/multidomain.py

from typing import Optional, Sequence, Tuple

import torch
import torch.nn.functional as F
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul


class MultiDomainConv2d(nn.Module):
    """Multi-domain convolution layer."""

    def __init__(
        self,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 1,  # pylint: disable=unused-argument
        in_channels: int = 4,
        out_channels: int = 4,
        **kwargs,
    ):
        """Inits :class:`MultiDomainConv2d`.

        Parameters
        ----------
        fft_centered : bool, optional
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str, optional
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        in_channels : int, optional
            Number of input channels. Default is ``4``.
        out_channels : int, optional
            Number of output channels. Default is ``4``.
        """
        super().__init__()

        self.image_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels // 2, **kwargs)
        self.kspace_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels // 2, **kwargs)

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = 1

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward method of :class:`MultiDomainConv2d`."""
        kspace = [
            fft2(im, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
            for im in torch.split(image.permute(0, 2, 3, 1).contiguous(), 2, -1)
        ]
        kspace = torch.cat(kspace, -1).permute(0, 3, 1, 2)
        kspace = self.kspace_conv(kspace)

        backward = [
            ifft2(
                ks.float(),
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            ).type(image.type())
            for ks in torch.split(kspace.permute(0, 2, 3, 1).contiguous(), 2, -1)  # type: ignore
        ]
        backward = torch.cat(backward, -1).permute(0, 3, 1, 2)
        image = self.image_conv(image)
        image = torch.cat([image, backward], dim=self.coil_dim)
        return image


class MultiDomainConvTranspose2d(nn.Module):
    """Multi-Domain convolutional transpose layer."""

    def __init__(
        self,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 1,  # pylint: disable=unused-argument
        in_channels: int = 4,
        out_channels: int = 4,
        **kwargs,
    ):
        """Inits :class:`MultiDomainConvTranspose2d`.

        Parameters
        ----------
        fft_centered : bool, optional
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str, optional
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        in_channels : int, optional
            Number of input channels. Default is ``4``.
        out_channels : int, optional
            Number of output channels. Default is ``4``.
        """
        super().__init__()
        self.image_conv = nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels // 2, **kwargs)
        self.kspace_conv = nn.ConvTranspose2d(in_channels=in_channels, out_channels=out_channels // 2, **kwargs)
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = 1

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward method of :class:`MultiDomainConvTranspose2d`."""
        kspace = [
            fft2(im, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
            for im in torch.split(image.permute(0, 2, 3, 1).contiguous(), 2, -1)
        ]
        kspace = torch.cat(kspace, -1).permute(0, 3, 1, 2)
        kspace = self.kspace_conv(kspace)

        backward = [
            ifft2(
                ks.float(),
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            ).type(image.type())
            for ks in torch.split(kspace.permute(0, 2, 3, 1).contiguous(), 2, -1)  # type: ignore
        ]
        backward = torch.cat(backward, -1).permute(0, 3, 1, 2)

        image = self.image_conv(image)
        return torch.cat([image, backward], dim=self.coil_dim)


class MultiDomainConvBlock(nn.Module):
    """A multi-domain convolutional block that consists of two multi-domain convolution layers each followed by
    instance normalization, LeakyReLU activation and dropout.
    """

    def __init__(
        self,
        fft_centered: bool = False,
        fft_normalization: str = "backwar",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 1,
        in_channels: int = 4,
        out_channels: int = 4,
        dropout_probability: float = 0.0,
    ):
        """Inits :class:`MultiDomainConvBlock`.

        Parameters
        ----------
        fft_centered : bool, optional
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str, optional
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        in_channels : int, optional
            Number of input channels. Default is ``4``.
        out_channels : int, optional
            Number of output channels. Default is ``4``.
        """
        super().__init__()

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.dropout_probability = dropout_probability

        self.layers = nn.Sequential(
            MultiDomainConv2d(
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
                self.coil_dim,
                in_channels,
                out_channels,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.InstanceNorm2d(out_channels),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(dropout_probability),
            MultiDomainConv2d(
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
                self.coil_dim,
                out_channels,
                out_channels,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.InstanceNorm2d(out_channels),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(dropout_probability),
        )

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`MultiDomainConvBlock`."""
        return self.layers(_input)

    def __repr__(self):
        """Representation of :class:`MultiDomainConvBlock`."""
        return (
            f"MultiDomainConvBlock(in_channels={self.in_channels}, out_channels={self.out_channels}, "
            f"dropout_probability={self.dropout_probability})"
        )


class TransposeMultiDomainConvBlock(nn.Module):
    """A Transpose Convolutional Block that consists of one convolution transpose layers followed by instance
    normalization and LeakyReLU activation.
    """

    def __init__(
        self,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Sequence[int] = None,
        coil_dim: int = 1,
        in_channels: int = 4,
        out_channels: int = 4,
    ):
        """Inits :class:`TransposeMultiDomainConvBlock`.

        Parameters
        ----------
        fft_centered : bool, optional
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str, optional
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        in_channels : int, optional
            Number of input channels. Default is ``4``.
        out_channels : int, optional
            Number of output channels. Default is ``4``.
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.layers = nn.Sequential(
            MultiDomainConvTranspose2d(
                fft_centered,
                fft_normalization,
                spatial_dims,
                coil_dim,
                in_channels,
                out_channels,
                kernel_size=2,
                stride=2,
                bias=False,
            ),
            nn.InstanceNorm2d(out_channels),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
        )

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        """Forward method of :class:`TransposeMultiDomainConvBlock`."""
        return self.layers(input_data)

    def __repr__(self):
        """Representation of :class:`TransposeMultiDomainConvBlock`."""
        return f"MultiDomainConvBlock(in_channels={self.in_channels}, out_channels={self.out_channels})"


class StandardizationLayer(nn.Module):
    r"""Multi-channel data standardization method. Inspired by AIRS model submission to the Fast MRI 2020 challenge.
    Given individual coil images :math:`'\'{x_i'\'}_{i=1}^{N_c}` and sensitivity coil maps
    :math:`'\'{S_i'\'}_{i=1}^{N_c}` it returns

    .. math::

        [(x_{sense}, {x_{res}}_1), ..., (x_{sense}, {x_{res}}_{N_c})]

    where

    :math:`{x_{res}}_i = xi - S_i X x_{sense}` and

    :math:`x_{sense} = '\'sum_{i=1}^{N_c} {S_i}^{*} X x_i`.
    """

    def __init__(self, coil_dim: int = 1, channel_dim: int = -1):
        """Inits :class:`StandardizationLayer`.

        Parameters
        ----------
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        channel_dim : int, optional
            Channel dimension. Default is ``-1``.
        """
        super().__init__()
        self.coil_dim = coil_dim
        self.channel_dim = channel_dim

    def forward(self, coil_images: torch.Tensor, sensitivity_map: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`StandardizationLayer`."""
        combined_image = complex_mul(coil_images, complex_conj(sensitivity_map)).sum(self.coil_dim)
        residual_image = combined_image.unsqueeze(self.coil_dim) - complex_mul(
            combined_image.unsqueeze(self.coil_dim), sensitivity_map
        )
        return torch.cat(
            [
                torch.cat(
                    [combined_image, residual_image.select(self.coil_dim, idx)],
                    self.channel_dim,
                ).unsqueeze(self.coil_dim)
                for idx in range(coil_images.size(self.coil_dim))
            ],
            self.coil_dim,
        )


class MultiDomainUnet2d(nn.Module):
    """Unet modification to be used with Multi-domain network as in AIRS Medical submission to the Fast MRI 2020
    challenge.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_filters: int,
        num_pool_layers: int,
        dropout_probability: float,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
    ):
        """Inits :class:`MultiDomainUnet2d`.

        Parameters
        ----------
        in_channels : int, optional
            Number of input channels.
        out_channels : int, optional
            Number of output channels.
        num_filters : int, optional
            Number of filters.
        num_pool_layers : int, optional
            Number of pooling layers.
        dropout_probability : float, optional
            Dropout probability.
        fft_centered : bool, optional
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str, optional
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        """
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_filters = num_filters
        self.num_pool_layers = num_pool_layers
        self.dropout_probability = dropout_probability
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        self.down_sample_layers = nn.ModuleList(
            [
                MultiDomainConvBlock(
                    self.fft_centered,
                    self.fft_normalization,
                    self.spatial_dims,
                    self.coil_dim,
                    in_channels,
                    num_filters,
                    dropout_probability,
                )
            ]
        )
        ch = num_filters
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers += [
                MultiDomainConvBlock(
                    self.fft_centered,
                    self.fft_normalization,
                    self.spatial_dims,
                    self.coil_dim,
                    ch,
                    ch * 2,
                    dropout_probability,
                )
            ]
            ch = ch * 2
        self.conv = MultiDomainConvBlock(
            self.fft_centered,
            self.fft_normalization,
            self.spatial_dims,
            self.coil_dim,
            ch,
            ch * 2,
            dropout_probability,
        )

        self.up_conv = nn.ModuleList()
        self.up_transpose_conv = nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_transpose_conv += [
                TransposeMultiDomainConvBlock(
                    self.fft_centered, self.fft_normalization, self.spatial_dims, self.coil_dim, ch * 2, ch
                )
            ]
            self.up_conv += [
                MultiDomainConvBlock(
                    self.fft_centered,
                    self.fft_normalization,
                    self.spatial_dims,
                    self.coil_dim,
                    ch * 2,
                    ch,
                    dropout_probability,
                )
            ]
            ch = ch // 2

        self.up_transpose_conv += [
            TransposeMultiDomainConvBlock(
                self.fft_centered, self.fft_normalization, self.spatial_dims, self.coil_dim, ch * 2, ch
            )
        ]
        self.up_conv += [
            nn.Sequential(
                MultiDomainConvBlock(
                    self.fft_centered,
                    self.fft_normalization,
                    self.spatial_dims,
                    self.coil_dim,
                    ch * 2,
                    ch,
                    dropout_probability,
                ),
                nn.Conv2d(ch, self.out_channels, kernel_size=1, stride=1),
            )
        ]

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`MultiDomainUnet2d`."""
        stack = []
        output = input_data

        # Apply down-sampling layers
        for layer in self.down_sample_layers:
            output = layer(output)
            stack.append(output)
            output = F.avg_pool2d(output, kernel_size=2, stride=2, padding=0)

        output = self.conv(output)

        # Apply up-sampling layers
        for transpose_conv, conv in zip(self.up_transpose_conv, self.up_conv):
            downsample_layer = stack.pop()
            output = transpose_conv(output)

            # Reflect pad on the right/bottom if needed to handle odd input dimensions.
            padding = [0, 0, 0, 0]
            if output.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # Padding right
            if output.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # Padding bottom
            if sum(padding) != 0:
                output = F.pad(output, padding, "reflect")

            output = torch.cat([output, downsample_layer], dim=1)
            output = conv(output)

        return output



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/classes/loss.py

import torch

from atommic.core.classes.common import Serialization, Typing

__all__ = ["Loss"]


class Loss(torch.nn.modules.loss._Loss, Typing, Serialization):  # pylint: disable=protected-access
    """Inherit this class to implement custom loss."""



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/exp_manager.py

import glob
import os
import subprocess
import sys
import time
import warnings
from dataclasses import dataclass
from pathlib import Path
from shutil import copy, move
from typing import Any, Dict, List, Optional, Tuple, Union

import pytorch_lightning
import torch
from hydra.core.hydra_config import HydraConfig
from hydra.utils import get_original_cwd
from omegaconf import DictConfig, OmegaConf, open_dict
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import Callback, ModelCheckpoint
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks.timer import Timer
from pytorch_lightning.loggers import TensorBoardLogger, WandbLogger
from pytorch_lightning.loops import _TrainingEpochLoop
from pytorch_lightning.strategies.ddp import DDPStrategy

import atommic.utils
from atommic.collections.common.callbacks import EMA
from atommic.constants import ATOMMIC_ENV_VARNAME_TESTING, ATOMMIC_ENV_VARNAME_VERSION
from atommic.utils import logging, timers
from atommic.utils.app_state import AppState
from atommic.utils.callbacks import ATOMMICModelCheckpoint, PreemptionCallback
from atommic.utils.env_var_parsing import get_envbool
from atommic.utils.exceptions import ATOMMICBaseException
from atommic.utils.get_rank import is_global_rank_zero
from atommic.utils.lightning_logger_patch import add_filehandlers_to_pl_logger


class NotFoundError(ATOMMICBaseException):
    """Raised when a file or folder is not found"""


class LoggerMisconfigurationError(ATOMMICBaseException):
    """Raised when a mismatch between trainer.logger and exp_manager occurs"""

    def __init__(self, message):
        """Inits :class:`LoggerMisconfigurationError`.

        Parameters
        ----------
        message : str
            The message to display.
        """
        message = (
            message + "You can disable lightning's trainer from creating a logger by passing logger=False to its "
            "constructor. "
        )
        super().__init__(message)


class CheckpointMisconfigurationError(ATOMMICBaseException):
    """Raised when a mismatch between trainer.callbacks and exp_manager occurs"""


@dataclass
class EarlyStoppingParams:
    """Parameters for the early stopping callback."""

    monitor: str = "val_loss"  # The metric that early stopping should consider.
    mode: str = "min"  # inform early stopping whether to look for increase or decrease in monitor.
    min_delta: float = 0.001  # smallest change to consider as improvement.
    patience: int = 10  # how many (continuous) validation cycles to wait with no improvement and stopping training.
    verbose: bool = True
    strict: bool = True
    check_finite: bool = True
    stopping_threshold: Optional[float] = None
    divergence_threshold: Optional[float] = None
    check_on_train_epoch_end: Optional[bool] = None
    log_rank_zero_only: bool = False


@dataclass
class CallbackParams:
    """Parameters for a callback"""

    filepath: Optional[str] = None  # Deprecated
    # If None, exp_manager will attempt to handle the filepath
    dirpath: Optional[str] = None
    # If None, exp_manager will attempt to handle the filepath
    filename: Optional[str] = None
    monitor: Optional[str] = "val_loss"
    verbose: Optional[bool] = True
    save_last: Optional[bool] = True
    save_top_k: Optional[int] = 3
    save_weights_only: Optional[bool] = False
    mode: Optional[str] = "min"
    auto_insert_metric_name: bool = True
    every_n_epochs: Optional[int] = 1
    every_n_train_steps: Optional[int] = None
    train_time_interval: Optional[str] = None
    # If None, exp_manager will attempt to handle the filepath
    prefix: Optional[str] = None
    postfix: str = ".atommic"
    save_best_model: bool = False
    always_save_atommic: bool = False
    # Automatically save .atommic file during on_train_end hook
    save_atommic_on_train_end: Optional[bool] = True
    # tensor parallel size * pipeline parallel size
    model_parallel_size: Optional[int] = None
    save_on_train_epoch_end: Optional[bool] = False  # Save after training, not after validation


@dataclass
class StepTimingParams:
    """Parameters for the step timing callback."""

    reduction: Optional[str] = "mean"
    # if True torch.cuda.synchronize() is called on start/stop
    sync_cuda: Optional[bool] = False
    # if positive, defines the size of a sliding window for computing mean
    buffer_size: Optional[int] = 1


@dataclass
class EMAParams:
    """Parameters for the EMA callback."""

    enable: Optional[bool] = False
    decay: Optional[float] = 0.999
    cpu_offload: Optional[bool] = False
    validate_original_weights: Optional[bool] = False
    every_n_steps: int = 1


@dataclass
class ExpManagerConfig:
    """Configuration for the experiment manager."""

    # Log dir creation parameters
    explicit_log_dir: Optional[str] = None
    exp_dir: Optional[str] = None
    name: Optional[str] = None
    version: Optional[str] = None
    use_datetime_version: Optional[bool] = True
    resume_if_exists: Optional[bool] = False
    resume_past_end: Optional[bool] = False
    resume_ignore_no_checkpoint: Optional[bool] = False
    resume_from_checkpoint: Optional[str] = None
    # Logging parameters
    create_tensorboard_logger: Optional[bool] = True
    summary_writer_kwargs: Optional[Dict[Any, Any]] = None
    create_wandb_logger: Optional[bool] = False
    wandb_logger_kwargs: Optional[Dict[Any, Any]] = None
    # Checkpointing parameters
    create_checkpoint_callback: Optional[bool] = True
    checkpoint_callback_params: Optional[CallbackParams] = CallbackParams()
    create_early_stopping_callback: Optional[bool] = False
    early_stopping_callback_params: Optional[EarlyStoppingParams] = EarlyStoppingParams()
    create_preemption_callback: Optional[bool] = True
    # Additional exp_manager arguments
    files_to_copy: Optional[List[str]] = None
    # logs timing of train/val/test steps
    log_step_timing: Optional[bool] = True
    step_timing_kwargs: Optional[StepTimingParams] = StepTimingParams()
    # Configures creation of log files for different ranks
    log_local_rank_0_only: Optional[bool] = False
    log_global_rank_0_only: Optional[bool] = False
    # disable initial validation when resuming from a checkpoint saved during validation
    disable_validation_on_resume: Optional[bool] = True
    ema: Optional[EMAParams] = EMAParams()
    # Wall clock time limit
    max_time_per_run: Optional[str] = None
    # time to sleep non 0 ranks during initialization
    seconds_to_sleep: float = 5


class TimingCallback(Callback):
    """Logs execution time of train/val/test steps"""

    def __init__(self, timer_kwargs=None):
        """Inits :class:`TimingCallback`."""
        if timer_kwargs is None:
            timer_kwargs = {}
        self.timer = timers.NamedTimer(**timer_kwargs)

    def _on_batch_start(self, name):
        """Called at the beginning of each batch"""
        # reset only if we do not return mean of a sliding window
        if self.timer.buffer_size <= 0:
            self.timer.reset(name)

        self.timer.start(name)

    def _on_batch_end(self, name, pl_module):
        """Called at the end of each batch"""
        self.timer.stop(name)
        # Set the `batch_size=1` as WAR for `dataloader_iter`, which is not used for any metric
        pl_module.log(
            name + ' in s',
            self.timer[name],
            on_step=True,
            on_epoch=False,
            batch_size=1,
            prog_bar=(name == "train_step_timing"),
        )

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, **kwargs):  # pylint: disable=unused-argument
        """Called at the beginning of each training batch"""
        self._on_batch_start("train_step_timing")

    def on_train_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, **kwargs  # pylint: disable=unused-argument
    ):
        """Logs the time taken by the training batch"""
        self._on_batch_end("train_step_timing", pl_module)

    def on_validation_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx=0):
        """Logs the time taken by the validation batch"""
        self._on_batch_start("validation_step_timing")

    def on_validation_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0):
        """Logs the time taken by the validation step"""
        self._on_batch_end("validation_step_timing", pl_module)

    def on_test_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx=0):
        """Logs execution time of test steps"""
        self._on_batch_start("test_step_timing")

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0):
        """Logs execution time of test steps"""
        self._on_batch_end("test_step_timing", pl_module)

    def on_before_backward(self, trainer, pl_module, loss):
        """Logs the time taken for backward pass"""
        self._on_batch_start("train_backward_timing")

    def on_after_backward(self, trainer, pl_module):
        """Note: this is called after the optimizer step"""
        self._on_batch_end("train_backward_timing", pl_module)


def exp_manager(trainer: Trainer, cfg: Optional[Union[DictConfig, Dict]] = None) -> Optional[Path]:  # noqa: MC0001
    r"""exp_manager is a helper function used to manage folders for experiments. It follows the pytorch lightning
    paradigm of exp_dir/model_or_experiment_name/version. If the lightning trainer has a logger, exp_manager will get
    exp_dir, name, and version from the logger. Otherwise, it will use the exp_dir and name arguments to create the
    logging directory. exp_manager also allows for explicit folder creation via explicit_log_dir.

    The version can be a datetime string or an integer. Date time version can be disabled if use_datetime_version is
    set to False. It optionally creates TensorBoardLogger, WandBLogger, ModelCheckpoint objects from pytorch lightning.
    It copies sys.argv, and git information if available to the logging directory. It creates a log file for each
    process to log their output into.

    exp_manager additionally has a resume feature (resume_if_exists) which can be used to continuing training from the
    constructed log_dir. When you need to continue the training repeatedly (like on a cluster which you need multiple
    consecutive jobs), you need to avoid creating the version folders. Therefore, from v1.0.0, when resume_if_exists
    is set to True, creating the version folders is ignored.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The lightning trainer object.
    cfg : DictConfig or Dict, optional
        Can have the following keys:
            - explicit_log_dir : str
                Can be used to override exp_dir/name/version folder creation. Defaults to ``None``, which will use
                    exp_dir, name, and version to construct the logging directory.
            - exp_dir : str
                The base directory to create the logging directory. Defaults to ``None``, which logs to
                    ./atommic_experiments.
            - name : str
                The name of the experiment. Defaults to ``None`` which turns into "default" via name = name or
                    "default".
            - version : str
                The version of the experiment. Defaults to None which uses either a datetime string or lightning's
                    TensorboardLogger system of using version_{int}.
            - use_datetime_version : bool
                Whether to use a datetime string for version. Default is ``True``.
            - resume_if_exists : bool
                Whether this experiment is resuming from a previous run. If True, it sets
                    trainer._checkpoint_connector._ckpt_path so that the trainer should auto-resume. exp_manager will
                    move files under log_dir to log_dir/run_{int}. Default is ``False``. When resume_if_exists is
                    True, we would not create version folders to make it easier to find the log folder for next runs.
            - resume_past_end : bool
                exp_manager errors out if resume_if_exists is True and a checkpoint matching '\'*end.ckpt indicating a
                    previous training run fully completed. This behaviour can be disabled, in which case the
                    '\'*end.ckpt will be loaded by setting resume_past_end to True. Default is ``False``.
            - resume_ignore_no_checkpoint : bool
                exp_manager errors out if resume_if_exists is True and no checkpoint could be found. This behaviour
                    can be disabled, in which case exp_manager will print a message and continue without restoring, by
                    setting resume_ignore_no_checkpoint to True. Default is ``False``.
            - resume_from_checkpoint : str
                Can be used to specify a path to a specific checkpoint file to load from. This will override any
                    checkpoint found when resume_if_exists is True. Default is ``None``.
            - create_tensorboard_logger : bool
                Whether to create a tensorboard logger and attach it to the pytorch lightning trainer.
                    Default is ``True``.
            - summary_writer_kwargs : dict
                A dictionary of kwargs that can be passed to lightning's TensorboardLogger class. Note that log_dir is
                    passed by exp_manager and cannot exist in this dict. Default is ``None``.
            - create_wandb_logger : bool
                Whether to create a Weights and Biases logger and attach it to the pytorch lightning  trainer.
                    Default is ``False``.
            - wandb_logger_kwargs : dict
                A dictionary of kwargs that can be passed to lightning's WandBLogger class. Note that name and project
                    are required parameters if create_wandb_logger is True. Default is ``None``..
            - create_checkpoint_callback : bool
                Whether to create a ModelCheckpoint callback and attach it to the pytorch lightning trainer. The
                    ModelCheckpoint saves the top 3 models with the best "val_loss", the most recent checkpoint under
                    '\'*last.ckpt, and the final checkpoint after training completes under '\'*end.ckpt.
                    Default is ``True``.
            - create_early_stopping_callback : bool
                Whether to create an EarlyStopping callback and attach it to the pytorch lightning trainer. The
                    EarlyStopping callback stops training if the "val_loss" does not improve for 3 epochs.
                    Default is ``True``.
            - files_to_copy : list
                A list of files to copy to the experiment logging directory. Defaults to None which copies no files.
            - log_local_rank_0_only : bool
                Whether to only create log files for local rank 0. Default is ``False``. Set this to True if you are
                    using DDP with many GPUs and do not want many log files in your exp dir.
            - log_global_rank_0_only : bool
                Whether to only create log files for global rank 0. Defaults to False. Set this to True if you are
                    using DDP with many GPUs and do not want many log files in your exp dir.
            - max_time : str
                The maximum wall clock time *per run*. This is intended to be used on clusters where you want a
                    checkpoint to be saved after this specified time and be able to resume from that checkpoint.
                    Default is ``None``.
            - seconds_to_sleep : float
                Seconds to sleep non rank 0 processes for. Used to give enough time for rank 0 to initialize.

    Returns
    -------
    log_dir : Path
        The final logging directory where logging files are saved. Usually the concatenation of exp_dir, name, and
        version.
    """
    # Add rank information to logger
    # Note: trainer.global_rank and trainer.is_global_zero are not set until trainer.fit, so have to hack around it
    local_rank = int(os.environ.get("LOCAL_RANK", 0))
    global_rank = trainer.node_rank * trainer.num_devices + local_rank
    logging.rank = global_rank

    if cfg is None:
        logging.error("exp_manager did not receive a cfg argument. It will be disabled.")
        return None

    if trainer.fast_dev_run:
        logging.info("Trainer was called with fast_dev_run. exp_manager will return without any functionality.")
        return None

    # Ensure passed cfg is compliant with ExpManagerConfig
    schema = OmegaConf.structured(ExpManagerConfig)
    if isinstance(cfg, dict):
        cfg = OmegaConf.create(cfg)
    elif not isinstance(cfg, DictConfig):
        raise ValueError(f"cfg was type: {type(cfg)}. Expected either a dict or a DictConfig")
    cfg = OmegaConf.create(OmegaConf.to_container(cfg, resolve=True))
    cfg = OmegaConf.merge(schema, cfg)

    # Ensures that trainer options are compliant with atommic and exp_manager arguments
    error_checks(trainer, cfg)

    log_dir, exp_dir, name, version = get_log_dir(
        trainer=trainer,
        exp_dir=cfg.exp_dir,
        name=cfg.name,
        version=cfg.version,
        explicit_log_dir=cfg.explicit_log_dir,
        use_datetime_version=cfg.use_datetime_version,
        resume_if_exists=cfg.resume_if_exists,
    )

    check_resume(
        trainer,
        log_dir,
        cfg.resume_if_exists,
        cfg.resume_past_end,
        cfg.resume_ignore_no_checkpoint,
        cfg.checkpoint_callback_params.dirpath,
        cfg.resume_from_checkpoint,
    )

    checkpoint_name = name
    # If name returned from get_log_dir is "", use cfg.name for checkpointing
    if checkpoint_name is None or checkpoint_name == "":
        checkpoint_name = cfg.name or "default"

    cfg.name = name  # Used for configure_loggers so that the log_dir is properly set even if name is ""
    cfg.version = version

    # update app_state with log_dir, exp_dir, etc
    app_state = AppState()
    app_state.log_dir = log_dir
    app_state.exp_dir = exp_dir
    app_state.name = name
    app_state.version = version
    app_state.checkpoint_name = checkpoint_name
    app_state.create_checkpoint_callback = cfg.create_checkpoint_callback
    app_state.checkpoint_callback_params = cfg.checkpoint_callback_params

    # Create the logging directory if it does not exist. Cannot limit creation to global zero as all ranks write to own
    # log file.
    os.makedirs(log_dir, exist_ok=True)
    logging.info(f"Experiments will be logged at {log_dir}")
    trainer._default_root_dir = log_dir  # pylint: disable=protected-access

    if cfg.log_local_rank_0_only is True and cfg.log_global_rank_0_only is True:
        raise ValueError(
            "Cannot set both log_local_rank_0_only and log_global_rank_0_only to True. "
            "Please set either one or neither."
        )

    # This is set if the env var atommic_TESTING is set to True.
    atommic_testing = get_envbool(ATOMMIC_ENV_VARNAME_TESTING, False)

    # Handle logging to file
    log_file = log_dir / f"atommic_log_globalrank-{global_rank}_localrank-{local_rank}.txt"
    if cfg.log_local_rank_0_only is True and not atommic_testing:
        if local_rank == 0:
            logging.add_file_handler(log_file)
    elif cfg.log_global_rank_0_only is True and not atommic_testing:
        if global_rank == 0:
            logging.add_file_handler(log_file)
    else:
        # Logs on all ranks.
        logging.add_file_handler(log_file)

    # For some reason, LearningRateLogger requires trainer to have a logger. Safer to create logger on all ranks
    # not just global rank 0.
    if cfg.create_tensorboard_logger or cfg.create_wandb_logger:
        configure_loggers(
            trainer,
            [Path(exp_dir)],
            [Path(log_dir)],
            cfg.name,
            cfg.version,
            cfg.checkpoint_callback_params,
            cfg.create_tensorboard_logger,
            cfg.summary_writer_kwargs,
            cfg.create_wandb_logger,
            cfg.wandb_logger_kwargs,
        )

    # add loggers timing callbacks
    if cfg.log_step_timing:
        timing_callback = TimingCallback(timer_kwargs=cfg.step_timing_kwargs or {})
        trainer.callbacks.insert(0, timing_callback)

    if cfg.ema.enable:
        ema_callback = EMA(
            decay=cfg.ema.decay,
            validate_original_weights=cfg.ema.validate_original_weights,
            cpu_offload=cfg.ema.cpu_offload,
            every_n_steps=cfg.ema.every_n_steps,
        )
        trainer.callbacks.append(ema_callback)

    if cfg.create_early_stopping_callback:
        early_stop_callback = EarlyStopping(**cfg.early_stopping_callback_params)
        trainer.callbacks.append(early_stop_callback)

    if cfg.create_checkpoint_callback:
        configure_checkpointing(
            trainer,
            log_dir,
            checkpoint_name,
            cfg.resume_if_exists,
            cfg.checkpoint_callback_params,
            cfg.create_preemption_callback,
        )

    if cfg.disable_validation_on_resume:
        # extend training loop to skip initial validation when resuming from checkpoint
        configure_no_restart_validation_training_loop(trainer)

    # Setup a stateless timer for use on clusters.
    if cfg.max_time_per_run is not None:
        found_ptl_timer = False
        for idx, callback in enumerate(trainer.callbacks):
            if isinstance(callback, Timer):
                # NOTE: PTL does not expose a `trainer.max_time`. By the time we are in this function, PTL has already
                # set up a timer if the user specifies `trainer.max_time` so best we can do is replace that.
                # Working: If only `trainer.max_time` is set - it behaves as a normal PTL timer.
                # If only `exp_manager.max_time_per_run` is set - it behaves as a StateLessTimer.
                # If both are set, it also behaves as a StateLessTimer.
                logging.warning(
                    "Found a PTL Timer callback, replacing with a StatelessTimer callback. "
                    "This will happen if you set trainer.max_time as well as exp_manager.max_time_per_run."
                )
                trainer.callbacks[idx] = StatelessTimer(cfg.max_time_per_run)
                found_ptl_timer = True
                break

        if not found_ptl_timer:
            trainer.max_time = cfg.max_time_per_run
            trainer.callbacks.append(StatelessTimer(cfg.max_time_per_run))

    if is_global_rank_zero():
        # Move files_to_copy to folder and add git information if present
        if cfg.files_to_copy:
            for _file in cfg.files_to_copy:
                copy(Path(_file), log_dir)

        # Create files for cmd args and git info
        with open(log_dir / "cmd-args.log", "w", encoding="utf-8") as _file:
            _file.write(" ".join(sys.argv))

        # Try to get git hash
        git_repo, git_hash = get_git_hash()
        if git_repo:
            with open(log_dir / "git-info.log", "w", encoding="utf-8") as _file:
                _file.write(f"commit hash: {git_hash}")
                _file.write(get_git_diff())

        # Add err_file logging to global_rank zero
        logging.add_err_file_handler(log_dir / "atommic_error_log.txt")

        # Add lightning file logging to global_rank zero
        add_filehandlers_to_pl_logger(log_dir / "lightning_logs.txt", log_dir / "atommic_error_log.txt")

    elif trainer.num_devices * trainer.num_devices > 1:
        # sleep other ranks so rank 0 can finish
        # doing the initialization such as moving files
        time.sleep(cfg.seconds_to_sleep)

    return log_dir


def error_checks(trainer: Trainer, cfg: Optional[Union[DictConfig, Dict]] = None):
    """Checks that the passed trainer is compliant with atommic and exp_manager's passed configuration. Checks that:
    - Throws error when hydra has changed the working directory. This causes issues with lightning's DDP
    - Throws error when trainer has loggers defined but create_tensorboard_logger or create_wandB_logger is True
    - Prints error messages when 1) run on multi-node and not Slurm, and 2) run on multi-gpu without DDP
    """
    if HydraConfig.initialized() and get_original_cwd() != os.getcwd():
        raise ValueError(
            "Hydra changed the working directory. This interferes with ExpManger's functionality. Please pass "
            "hydra.run.dir=. to your python script."
        )

    if trainer.logger is not None and (cfg.create_tensorboard_logger or cfg.create_wandb_logger):  # type: ignore
        raise LoggerMisconfigurationError(
            "The pytorch lightning trainer that was passed to exp_manager contained a logger, and either "
            f"create_tensorboard_logger: {cfg.create_tensorboard_logger} or create_wandb_logger: "  # type: ignore
            f"was set to True. These can only be used if trainer does not already have a logger."
        )

    if trainer.num_nodes > 1 and not check_slurm(trainer):
        logging.error(
            "You are running multi-node training without SLURM handling the processes."
            " Please note that this is not tested in atommic and could result in errors."
        )

    if trainer.num_devices > 1 and not isinstance(trainer.strategy, DDPStrategy):
        logging.error(
            "You are running multi-gpu without ddp.Please note that this is not tested in atommic and could result in "
            "errors."
        )


def check_resume(  # noqa: MC0001
    trainer: Trainer,
    log_dir: Union[str, Path],
    resume_if_exists: bool = False,
    resume_past_end: bool = False,
    resume_ignore_no_checkpoint: bool = False,
    dirpath: str = None,
    resume_from_checkpoint: str = None,
):
    """Checks that resume=True was used correctly with the arguments pass to exp_manager. Sets
    trainer._checkpoint_connector._ckpt_path as necessary.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The trainer that is being used.
    log_dir : Union[str, Path]
        The directory where the logs are being saved.
    resume_if_exists : bool
        Whether to resume if the experiment directory already exists.
    resume_past_end : bool
        Whether to resume from the end of the experiment.
    resume_ignore_no_checkpoint : bool
        Whether to ignore if there is no checkpoint to resume from.
    dirpath : str
        The directory to resume from. If None, will resume from the latest checkpoint.
    resume_from_checkpoint : str
        The checkpoint to resume from. If None, will resume from the latest checkpoint.

    Returns
    -------
    NotFoundError : bool
        If resume is True, resume_ignore_no_checkpoint is False, and checkpoints could not be found.
    ValueError : bool
        If resume is True, and there were more than 1 checkpoint could be found.
    """
    if not log_dir:
        raise ValueError(f"Resuming requires the log_dir {log_dir} to be passed to exp_manager")

    checkpoint = None
    if resume_from_checkpoint:
        checkpoint = resume_from_checkpoint
    if resume_if_exists:
        # Use <log_dir>/checkpoints/ unless `dirpath` is set
        checkpoint_dir = Path(dirpath) if dirpath else Path(Path(log_dir) / "checkpoints")

        # when using distributed checkpointing, checkpoint_dir is a directory of directories
        # we check for this here
        dist_checkpoints = [d for d in list(checkpoint_dir.glob("*")) if d.is_dir()]
        end_dist_checkpoints = [d for d in dist_checkpoints if d.match("*end")]
        last_dist_checkpoints = [d for d in dist_checkpoints if d.match("*last")]

        end_checkpoints = end_dist_checkpoints if end_dist_checkpoints else list(checkpoint_dir.rglob("*end.ckpt"))
        last_checkpoints = last_dist_checkpoints if last_dist_checkpoints else list(checkpoint_dir.rglob("*last.ckpt"))

        if not checkpoint_dir.exists() or (not len(end_checkpoints) > 0 and not len(last_checkpoints) > 0):
            if resume_ignore_no_checkpoint:
                warn = (
                    "There were no checkpoints found in checkpoint_dir or no checkpoint folder at checkpoint_dir "
                    f":{checkpoint_dir}. "
                )
                if checkpoint is None:
                    warn += "Training from scratch."
                elif checkpoint == resume_from_checkpoint:
                    warn += f"Training from {resume_from_checkpoint}."
                logging.warning(warn)
            else:
                raise NotFoundError(
                    "There were no checkpoints found in checkpoint_dir or no checkpoint folder at checkpoint_dir "
                    f":{checkpoint_dir}. Cannot resume."
                )
        elif len(end_checkpoints) > 0:
            if resume_past_end:
                if len(end_checkpoints) > 1:
                    if 'mp_rank' in str(end_checkpoints[0]):
                        checkpoint = end_checkpoints[0]  # type: ignore
                    else:
                        raise ValueError(f"Multiple checkpoints {end_checkpoints} that matches *end.ckpt.")
            else:
                raise ValueError(
                    f"Found {end_checkpoints[0]} indicating that the last training run has already completed."
                )
        elif len(last_checkpoints) > 1:
            if 'mp_rank' in str(last_checkpoints[0]) or 'tp_rank' in str(last_checkpoints[0]):
                checkpoint = last_checkpoints[0]  # type: ignore
                checkpoint = atommic.utils.model_utils.uninject_model_parallel_rank(checkpoint)
            else:
                raise ValueError(f"Multiple checkpoints {last_checkpoints} that matches *last.ckpt.")
        else:
            checkpoint = last_checkpoints[0]  # type: ignore

    # PTL 2.0 supports ckpt_path instead of resume_from_checkpoint as the trainer flag
    if checkpoint is not None:
        trainer.ckpt_path = str(checkpoint)
        logging.info(f'Resuming training from checkpoint: {trainer.ckpt_path}')

    if is_global_rank_zero():
        # Check to see if any files exist that need to be moved
        files_to_move = []
        if Path(log_dir).exists():
            for child in Path(log_dir).iterdir():
                if child.is_file():
                    files_to_move.append(child)

        if len(files_to_move) > 0:
            # Move old files to a new folder
            other_run_dirs = Path(log_dir).glob("run_*")
            run_count = 0
            for fold in other_run_dirs:
                if fold.is_dir():
                    run_count += 1
            new_run_dir = Path(Path(log_dir) / f"run_{run_count}")
            new_run_dir.mkdir()
            for _file in files_to_move:
                move(str(_file), str(new_run_dir))


def check_explicit_log_dir(
    trainer: Trainer,
    explicit_log_dir: List[Union[Path, str]],
    exp_dir: str,
    name: str,  # pylint: disable=unused-argument
    version: str,
) -> Tuple[Path, str, str, str]:
    """Checks that the passed arguments are compatible with explicit_log_dir.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The trainer to check.
    explicit_log_dir : str
        The explicit log dir to check.
    exp_dir : str
        The experiment directory to check.
    name : str
        The experiment name to check.
    version : str
        The experiment version to check.

    Returns
    -------
    tuple
        The log_dir, exp_dir, name, and version that should be used.

    Raises
    ------
    LoggerMisconfigurationError
        If the trainer already has a logger.
    """
    if trainer.logger is not None:
        raise LoggerMisconfigurationError(
            "The pytorch lightning trainer that was passed to exp_manager contained a logger and explicit_log_dir: "
            f"{explicit_log_dir} was pass to exp_manager. Please remove the logger from the lightning trainer."
        )
    # Checking only (explicit_log_dir) vs (exp_dir and version).
    # The `name` will be used as the actual name of checkpoint/archive.
    if exp_dir or version:
        logging.error(
            f"exp_manager received explicit_log_dir: {explicit_log_dir} and at least one of exp_dir: {exp_dir}, "
            f"or version: {version}. Please note that exp_dir, name, and version will be ignored."
        )
    if is_global_rank_zero() and Path(str(explicit_log_dir)).exists():
        logging.warning(f"Exp_manager is logging to {explicit_log_dir}, but it already exists.")
    return Path(str(explicit_log_dir)), str(explicit_log_dir), "", ""


def get_log_dir(
    trainer: Trainer,
    exp_dir: str = None,
    name: str = None,
    version: str = None,
    explicit_log_dir: str = None,
    use_datetime_version: bool = True,
    resume_if_exists: bool = False,
) -> Tuple[Path, str, str, str]:
    """Obtains the log_dir used for exp_manager.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The trainer to check.
    exp_dir : str
        The experiment directory to check.
    name : str
        The experiment name to check.
    version : str
        The experiment version to check.
    explicit_log_dir : str
        The explicit log dir to check.
    use_datetime_version : bool
        Whether to use datetime versioning.
    resume_if_exists : bool
        Whether to resume if the log_dir already exists.

    Raises
    -------
    LoggerMisconfigurationError : bool
        If trainer is incompatible with arguments.
    NotFoundError : bool
        If resume is True, resume_ignore_no_checkpoint is False, and checkpoints could not be found.
    ValueError : bool
        If resume is True, and there were more than 1 checkpoint could be found.
    """
    if explicit_log_dir:  # If explicit log_dir was passed, short circuit
        return check_explicit_log_dir(trainer, [Path(explicit_log_dir)], exp_dir, name, version)  # type: ignore

    # Default exp_dir to ./atommic_experiments if None was passed
    _exp_dir = exp_dir
    if exp_dir is None:
        _exp_dir = str(Path.cwd() / "atommic_experiments")

    # If the user has already defined a logger for the trainer, use the logger defaults for logging directory
    if trainer.logger is not None:
        if trainer.logger.save_dir:
            if exp_dir:
                raise LoggerMisconfigurationError(
                    "The pytorch lightning trainer that was passed to exp_manager contained a logger, the logger's "
                    f"save_dir was not None, and exp_dir ({exp_dir}) was not None. If trainer.logger.save_dir "
                    "exists, exp_manager will use trainer.logger.save_dir as the logging directory and exp_dir "
                    "must be None."
                )
            _exp_dir = trainer.logger.save_dir
        if name:
            raise LoggerMisconfigurationError(
                "The pytorch lightning trainer that was passed to exp_manager contained a logger, and name: "
                f"{name} was also passed to exp_manager. If the trainer contains a "
                "logger, exp_manager will use trainer.logger.name, and name passed to exp_manager must be None."
            )
        name = trainer.logger.name
        version = f"version_{trainer.logger.version}"
    # Use user-defined exp_dir, project_name, exp_name, and versioning options
    else:
        name = name or "default"
        version = version or os.environ.get(ATOMMIC_ENV_VARNAME_VERSION)

        if not version:
            if resume_if_exists:
                logging.warning(
                    "No version folders would be created under the log folder as 'resume_if_exists' is enabled."
                )
                version = None
            elif is_global_rank_zero():
                if use_datetime_version:
                    version = time.strftime("%Y-%m-%d_%H-%M-%S")
                else:
                    tensorboard_logger = TensorBoardLogger(save_dir=_exp_dir, name=name, version=version)
                    version = f"version_{tensorboard_logger.version}"
                os.environ[ATOMMIC_ENV_VARNAME_VERSION] = "" if version is None else version

    log_dir = Path(str(_exp_dir)) / Path(str(name)) / Path("" if version is None else str(version))
    return log_dir, str(_exp_dir), str(name), str(version)


def get_git_hash():
    """Helper function that tries to get the commit hash if running inside a git folder.

    Returns
    -------
    Bool: Whether the git subprocess ran without error.
    String: git subprocess output or error message
    """
    try:
        return True, subprocess.check_output(["git", "rev-parse", "HEAD"], stderr=subprocess.STDOUT).decode()
    except subprocess.CalledProcessError as err:
        return False, f'{err.output.decode("utf-8")}\n'


def get_git_diff():
    """Helper function that tries to get the git diff if running inside a git folder.

    Returns
    -------
    bool
        Whether the git subprocess ran without error.
    str
        If git subprocess output or error message.
    """
    try:
        return subprocess.check_output(["git", "diff"], stderr=subprocess.STDOUT).decode()
    except subprocess.CalledProcessError as err:
        return f'{err.output.decode("utf-8")}\n'


def configure_loggers(
    trainer: Trainer,
    exp_dir: List[Union[Path, str]],
    log_dir: List[Union[Path, str]],  # pylint: disable=unused-argument
    name: str,
    version: str,
    checkpoint_callback_params: dict,  # pylint: disable=unused-argument
    create_tensorboard_logger: bool,
    summary_writer_kwargs: dict,
    create_wandb_logger: bool,
    wandb_kwargs: dict,
):
    """Creates TensorboardLogger and/or WandBLogger and attach them to trainer. Raises ValueError if
    summary_writer_kwargs or wandb_kwargs are miss configured.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The trainer to attach the loggers to.
    exp_dir : str
        The experiment directory.
    log_dir : str
        The logging directory.
    name : str
        The name of the experiment.
    version : str
        The version of the experiment.
    checkpoint_callback_params : dict
        The checkpoint callback parameters.
    create_tensorboard_logger : bool
        Whether to create a TensorboardLogger.
    summary_writer_kwargs : dict
        The kwargs to pass to the TensorboardLogger.
    create_wandb_logger : bool
        Whether to create a Weights & Biases logger.
    wandb_kwargs : dict
        The kwargs to pass to the Weights & Biases logger.

    Returns
    -------
    LoggerList
        A list of loggers.
    """
    # Potentially create tensorboard logger and/or WandBLogger
    logger_list = []
    if create_tensorboard_logger:
        if summary_writer_kwargs is None:
            summary_writer_kwargs = {}
        elif "log_dir" in summary_writer_kwargs:
            raise ValueError(
                "You cannot pass `log_dir` as part of `summary_writer_kwargs`. `log_dir` is handled by lightning's "
                "TensorBoardLogger logger."
            )
        tensorboard_logger = TensorBoardLogger(
            save_dir=exp_dir[0], name=name, version=version, **summary_writer_kwargs
        )
        logger_list.append(tensorboard_logger)
        logging.info("TensorboardLogger has been set up")

    if create_wandb_logger:
        if wandb_kwargs is None:
            wandb_kwargs = {}
        if "name" not in wandb_kwargs and "project" not in wandb_kwargs:
            raise ValueError("name and project are required for wandb_logger")
        wandb_logger = WandbLogger(save_dir=str(exp_dir[0]), version=version, **wandb_kwargs)

        logger_list.append(wandb_logger)
        logging.info("WandBLogger has been set up")

    trainer._logger_connector.configure_logger(logger_list)  # pylint: disable=protected-access


def configure_checkpointing(  # noqa: MC0001
    trainer: Trainer,
    log_dir: Path,
    name: str,
    resume: bool,
    params: "DictConfig",
    create_preemption_callback: bool,
):
    """Adds ModelCheckpoint to trainer. Raises CheckpointMisconfigurationError if trainer already has a ModelCheckpoint
    callback or if trainer.weights_save_path was passed to Trainer.
    """
    for callback in trainer.callbacks:
        if isinstance(callback, ModelCheckpoint):
            raise CheckpointMisconfigurationError(
                "The pytorch lightning trainer that was passed to exp_manager contained a ModelCheckpoint "
                "and create_checkpoint_callback was set to True. Please either set create_checkpoint_callback "
                "to False, or remove ModelCheckpoint from the lightning trainer"
            )

    # Create the callback and attach it to trainer
    if "filepath" in params:
        if params.filepath is not None:
            logging.warning("filepath is deprecated. Please switch to dirpath and filename instead")
            if params.dirpath is None:
                params.dirpath = Path(params.filepath).parent
            if params.filename is None:
                params.filename = Path(params.filepath).name
        with open_dict(params):
            del params["filepath"]
    if params.dirpath is None:
        params.dirpath = Path(log_dir / "checkpoints")
    if params.filename is None:
        params.filename = f"{name}--{{{params.monitor}:.4f}}-{{epoch}}"
    if params.prefix is None:
        params.prefix = name
    ATOMMICModelCheckpoint.CHECKPOINT_NAME_LAST = f"{params.filename}-last"

    logging.debug(params.dirpath)
    logging.debug(params.filename)
    logging.debug(params.prefix)

    if "val" in params.monitor:
        if (
            trainer.max_epochs is not None
            and trainer.max_epochs != -1
            and trainer.max_epochs < trainer.check_val_every_n_epoch
        ):
            logging.error(
                "The checkpoint callback was told to monitor a validation value but trainer.max_epochs("
                f"{trainer.max_epochs}) was less than trainer.check_val_every_n_epoch("
                f"{trainer.check_val_every_n_epoch}). It is very likely this run will fail with "
                f"ModelCheckpoint(monitor='{params.monitor}') not found in the returned metrics. Please ensure that "
                "validation is run within trainer.max_epochs."
            )
        elif trainer.max_steps is not None and trainer.max_steps != -1:
            logging.warning(
                "The checkpoint callback was told to monitor a validation value and trainer's max_steps was set to "
                f"{trainer.max_steps}. Please ensure that max_steps will run for at least "
                f"{trainer.check_val_every_n_epoch} epochs to ensure that checkpointing will not error out."
            )

    checkpoint_callback = ATOMMICModelCheckpoint(n_resume=resume, **params)
    checkpoint_callback.last_model_path = trainer.ckpt_path or ""
    if "mp_rank" in checkpoint_callback.last_model_path or "tp_rank" in checkpoint_callback.last_model_path:
        checkpoint_callback.last_model_path = atommic.utils.model_utils.uninject_model_parallel_rank(
            checkpoint_callback.last_model_path
        )
    trainer.callbacks.append(checkpoint_callback)
    if create_preemption_callback:
        # Check if cuda is available as preemption is supported only on GPUs
        if torch.cuda.is_available():
            # By default, PreemptionCallback handles SIGTERM.
            # To handle other signals pass the signal in the call as below:
            # PreemptionCallback(checkpoint_callback, signal.SIGCHLD)
            preemption_callback = PreemptionCallback(checkpoint_callback)
            trainer.callbacks.append(preemption_callback)
        else:
            logging.info("Preemption is supported only on GPUs, disabling preemption")


def check_slurm(trainer):
    """Checks if the trainer is running on a slurm cluster. If so, it will check if the trainer is running on the
    master node. If it is not, it will exit.

    Parameters
    ----------
    trainer : pytorch_lightning.Trainer
        The trainer to check.

    Returns
    -------
    bool
        True if the trainer is running on the master node, False otherwise.
    """
    try:
        return trainer.accelerator_connector.is_slurm_managing_tasks
    except AttributeError:
        return False


class StatelessTimer(Timer):
    """Extension of PTL timers to be per run."""

    # pylint: disable=arguments-differ
    @staticmethod
    def state_dict() -> Dict[str, Any]:
        """Saves the state of the timer."""
        return {}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Loads the state of the timer."""


def configure_no_restart_validation_training_loop(trainer: pytorch_lightning.Trainer) -> None:
    """Configure the training loop to skip validation when resuming from a checkpoint."""
    if not isinstance(trainer.fit_loop.epoch_loop, _TrainingEpochLoop):
        warnings.warn("Detected custom epoch loop. Skipping no validation on restart support.", UserWarning)
        return
    # Pass trainer object to avoid trainer getting overwritten as None
    loop = SkipResumeTrainingValidationLoop(trainer, trainer.min_steps, trainer.max_steps)
    trainer.fit_loop.epoch_loop = loop


class SkipResumeTrainingValidationLoop(_TrainingEpochLoop):
    """Extend the PTL Epoch loop to skip validating when resuming. This happens when resuming a checkpoint that has
    already run validation, but loading restores the training state before validation has run.
    """

    def _should_check_val_fx(self) -> bool:
        """Skip validation if we are resuming from a checkpoint and the global step is a multiple of the validation."""
        if self.restarting and self.global_step % self.trainer.val_check_batch == 0:
            return False
        return super()._should_check_val_fx()


def clean_exp_ckpt(exp_log_dir: Union[str, Path], remove_ckpt: bool = True, remove_atommic: bool = False):
    """Helper method that removes Pytorch Lightning .ckpt files or atommic .atommic files from the checkpoint
    directory.

    Parameters
    ----------
    exp_log_dir : str or Path
        Path to the root directory of the current experiment.
    remove_ckpt : bool, optional
        Whether to remove all *.ckpt files in the checkpoints directory. Default is True.
    remove_atommic : bool, optional
        Whether to remove all *.atommic files in the checkpoints directory. Default is False.
    """
    exp_log_dir = str(exp_log_dir)

    if remove_ckpt:
        logging.info("Deleting *.ckpt files ...")
        ckpt_files = glob.glob(os.path.join(exp_log_dir, "checkpoints", "*.ckpt"))
        for filepath in ckpt_files:
            os.remove(filepath)
            logging.info(f"Deleted file : {filepath}")

    if remove_atommic:
        logging.info("Deleting *.atommic files ...")
        atommic_files = glob.glob(os.path.join(exp_log_dir, "checkpoints", "*.atommic"))
        for filepath in atommic_files:
            os.remove(filepath)
            logging.info(f"Deleted file : {filepath}")



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from torch import nn


class Conv3dBlock(nn.Module):
    """A 3D convolutional block."""

    def __init__(self, in_chans: int, out_chans: int, drop_prob: float, **kwargs):
        """Inits :class:`Conv3dBlock`.

        Parameters
        ----------
        in_chans : int
            Number of input channels.
        out_chans : int
            Number of output channels.
        drop_prob : float
            Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        self.layers = nn.Sequential(
            nn.Conv3d(in_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm3d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout3d(drop_prob),
            nn.Conv3d(out_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm3d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout3d(drop_prob),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`Conv3dBlock`."""
        return self.layers(image)


class TransposeConv3dBlock(nn.Module):
    """A 3D transposed convolutional block."""

    def __init__(self, in_chans: int, out_chans: int):
        """Inits :class:`TransposeConv3dBlock`.

        Parameters
        ----------
        in_chans : int
            Number of input channels.
        out_chans : int
            Number of output channels.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans

        self.layers = nn.Sequential(
            nn.ConvTranspose3d(in_chans, out_chans, kernel_size=(1, 2, 2), stride=(1, 2, 2), bias=False),
            nn.InstanceNorm3d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`TransposeConv3dBlock`."""
        return self.layers(image)


class UNet3D(nn.Module):
    """Implementation of the (3D) UNet for MRI segmentation, as presented in [Ronneberger2015]_.

    References
    ----------
    .. [Ronneberger2015] O. Ronneberger, P. Fischer, and Thomas Brox. U-net: Convolutional networks for biomedical
        image segmentation. In International Conference on Medical image computing and computer-assisted intervention,
        pages 234–241. Springer, 2015.
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        chans: int = 32,
        num_pool_layers: int = 4,
        drop_prob: float = 0.0,
        block=Conv3dBlock,
        **kwargs,
    ):
        """Inits :class:`UNet3D`.

        Parameters
        ----------
        in_chans : int
            Number of input channels.
        out_chans : int
            Number of output channels.
        chans : int
            Number of output channels of the first convolutional layer. Default is ``32``.
        num_pool_layers : int
            Number of down-sampling and up-sampling layers. Default is ``4``.
        drop_prob : float
            Dropout probability. Default is ``0.0``.
        block : nn.Module
            Convolutional block to use. Default is ``Conv3dBlock``.
        """
        super().__init__()
        self.in_chans = in_chans
        self.out_chans = out_chans
        self.chans = chans
        self.num_pool_layers = num_pool_layers
        self.drop_prob = drop_prob

        self.down_sample_layers = nn.ModuleList([Conv3dBlock(in_chans, chans, drop_prob)])
        ch = chans
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers.append(block(ch, ch * 2, drop_prob, **kwargs))
            ch = ch * 2
        self.conv = block(ch, ch * 2, drop_prob, **kwargs)

        self.up_conv = nn.ModuleList()
        self.up_transpose_conv = nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_transpose_conv.append(TransposeConv3dBlock(ch * 2, ch))
            self.up_conv.append(Conv3dBlock(ch * 2, ch, drop_prob))
            ch = ch // 2

        self.up_transpose_conv.append(TransposeConv3dBlock(ch * 2, ch))
        self.up_conv.append(
            nn.Sequential(
                Conv3dBlock(ch * 2, ch, drop_prob, **kwargs),
                nn.Conv3d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of  :class:`UNet3D`."""
        stack = []
        output = image

        # apply down-sampling layers
        for layer in self.down_sample_layers:
            output = layer(output)
            stack.append(output)
            output = nn.functional.avg_pool3d(output, kernel_size=(1, 2, 2), stride=(1, 2, 2), padding=0)

        output = self.conv(output)

        # apply up-sampling layers
        for transpose_conv, conv in zip(self.up_transpose_conv, self.up_conv):
            downsample_layer = stack.pop()
            output = transpose_conv(output)

            # reflect pad on the right/bottom if needed to handle odd input dimensions
            padding = [0, 0, 0, 0, 0, 0]
            if output.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # padding right
            if output.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # padding bottom
            if output.shape[-3] != downsample_layer.shape[-3]:
                padding[5] = 1  # padding back
            if torch.sum(torch.tensor(padding)) != 0:
                output = nn.functional.pad(output, padding, "reflect")

            output = torch.cat([output, downsample_layer], dim=1)
            output = conv(output)

        return output



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.collections.reconstruction.nn.varnet_base.varnet_block import VarNetBlock
from atommic.core.classes.common import typecheck

__all__ = ["VarNet"]


class VarNet(BaseMRIReconstructionModel):
    """Implementation of the End-to-end Variational Network (VN), as presented in [Sriram2020]_.

    References
    ----------
    .. [Sriram2020] Sriram A, Zbontar J, Murrell T, Defazio A, Zitnick CL, Yakubova N, Knoll F, Johnson P. End-to-end
        variational networks for accelerated MRI reconstruction. InInternational Conference on Medical Image Computing
        and Computer-Assisted Intervention 2020 Oct 4 (pp. 64-73). Springer, Cham.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`VarNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.no_dc = cfg_dict.get("no_dc")
        self.num_cascades = cfg_dict.get("num_cascades")

        # Cascades of VN blocks
        self.cascades = torch.nn.ModuleList(
            [
                VarNetBlock(
                    NormUnet(
                        chans=cfg_dict.get("channels", 18),
                        num_pools=cfg_dict.get("pooling_layers", 4),
                        in_chans=cfg_dict.get("in_chans", 2),
                        out_chans=cfg_dict.get("out_chans", 2),
                        drop_prob=cfg_dict.get("dropout", 0.0),
                        padding_size=cfg_dict.get("padding_size", 11),
                        normalize=cfg_dict.get("normalize", True),
                    ),
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                    no_dc=self.no_dc,
                )
                for _ in range(self.num_cascades)
            ]
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`VarNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        prediction = y.clone()
        for cascade in self.cascades:
            prediction = cascade(prediction, y, sensitivity_maps, mask)
        return check_stacked_complex(
            coil_combination_method(
                ifft2(prediction, self.fft_centered, self.fft_normalization, self.spatial_dims),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )
